// -*- mode:c++ -*-
//
// Module clamp_support.cc
//
// Copyright (c) 2002 Raoul M. Gough
//
// This material is provided "as is", with absolutely no warranty expressed
// or implied. Any use is at your own risk.
//
// Permission to use or copy this software for non-commercial purposes
// is hereby granted without fee, provided the above notices are
// retained on all copies.  Permission to modify the code and to
// distribute modified code is granted, provided the above notices are
// retained, and a notice that the code was modified is included with
// the above copyright notice.
//
// History
// =======
// 2002/ 2/10	rmg	File creation
//
// version: $Id: clamp_support.cc,v 1.11 2002/09/16 17:09:45 Raoul Exp $
//

#include "clamp_support.hh"
#include <iostream>
// #include <ostream>

using namespace std;

namespace clamp
{
  lambdaT::lambdaT ()
    : ret_type (0)
    , context (0)
    , params (0)
    , explicitContext (false)
    , onHeap (false)
  {
  }

  template<typename T>
  void safeAssign (auto_ptr<T> &to, T *from, bool force = false)
  {
    if (from || force)
      {
	to.reset (from);
      }
  }

  lambdaT *makelambda (lambdaT *lamp
		       , stringT *typestr
		       , paramListT *ctx
		       , paramListT *params
		       , stringT *qualstr
		       , bool onh)
  {
    if (lamp == 0)
      {
	lamp = new lambdaT;
      }
      
    //fprintf(stderr, "Creating lambda (%s)\n", qualstr ? qualstr->c_str() : "");

    safeAssign (lamp->ret_type, typestr);
    safeAssign (lamp->context, ctx);
    safeAssign (lamp->params, params);
    safeAssign (lamp->qual_str, qualstr);
    lamp->onHeap = onh;

    return lamp;
  }

  lambdaT *shiftlambda (lambdaT *lamp
			, paramListT *params)
  {
    // Release should be safe because it is only called if lamp is
    // already non-null, in which case makelambda guarantees no-throw
    lambdaT *result
      = makelambda (lamp, 0, lamp ? lamp->params.release() : 0, params,
                    lamp->qual_str.get());

    result->explicitContext = true;

    return result;
  }

  paramListT *appendplist (paramListT *front, paramListT *back)
  {
    if (front)
      {
	if (back)
	  {
	    front->splice (front->end(), *back);
	    delete back;
	  }
      }

    else if (back)
      {
	front = back;
      }

    return front;
  }

  paramListT *makeplist (stringT *type, stringT *name, stringT *i)
  {
    auto_ptr<paramListT> listp (new paramListT);

    listp->push_back (boost::shared_ptr<paramT> (new paramT (type, name, i)));

    return listp.release();
  }

  bool plistInitialisers (const paramListT *list)
  {
    bool foundInit = false;

    if (list)
      {
	for (paramListT::const_iterator iter = list->begin()
	       ; (iter != list->end()) && (!foundInit)
	       ; ++iter)
	  {
	    foundInit = (*iter)->initialiser.get();
	  }
      }

    return foundInit;
  }

  stringT *appendstr (stringT *s1, stringT *s2)
  {
    if (s1)
      {
	if (s2)
	  {
	    s1->append (" ");
	    s1->append (*s2);

	    delete s2;
	  }
      }

    else if (s2)
      {
	s1 = s2;
      }

    return s1;
  }

  stringT *appendstr (stringT *s1, stringT *s2, stringT *s3)
  {
    return appendstr (appendstr (s1, s2), s3);
  }

  stringT *appendstr (stringT *s1, stringT *s2, stringT *s3, stringT *s4)
  {
    return appendstr (appendstr (s1, s2, s3), s4);
  }

  stringT *appendstr (stringT *s1
		      , stringT *s2
		      , stringT *s3
		      , stringT *s4
		      , stringT *s5)
  {
    return appendstr (appendstr (s1, s2, s3, s4), s5);
  }

  stringT *makestr (const char *cstr)
  {
    if (cstr)
      {
	return new stringT (cstr);
      }

    else
      {
	return 0;
      }
  }

  stringT *makestr (const char *cstr, int length)
  {
    if (cstr)
      {
	return new stringT (cstr, length);
      }

    else
      {
	return 0;
      }
  }

  std::ostream &operator<< (std::ostream &strm, const lambdaT &lam)
  {
    strm
      << "lambdaT returning "
      << lam.ret_type.get()
      << " from";

    if (lam.explicitContext)
      {
	strm
	  << " context "
	  << lam.context.get()
	  << " and";
      }

    strm
      << " params "
      << lam.params.get();

    return strm;
  }

  std::ostream &operator<< (std::ostream &strm, const paramT &p)
  {
    return strm
      << "(type "
      << p.type.get()
      << " with name "
      << p.name.get()
      << ")";
  }

  std::ostream &operator<< (std::ostream &strm, const paramListT &pList)
  {
    for (paramListT::const_iterator iter = pList.begin()
	   ; iter != pList.end()
	   ; ++iter)
      {
	if (iter != pList.begin())
	  {
	    strm << ", ";
	  }

	strm << iter->get();
      }

    return strm;
  }
}
